/***************************************************************************
 *   Copyright (C) 2004 by Robin Atwood                                    *
 *   robin.atwood@attglobal.net                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "kps.h"
#include "renice.h"

#include <qpainter.h>
#include <qlabel.h>
#include <q3paintdevicemetrics.h>
#include <q3whatsthis.h>
#include <qregexp.h>
#include <qdebug.h>
//Added by qt3to4:
#include <QPixmap>

#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdeversion.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <kconfig.h>
#include <kmessagebox.h>
#include <kedittoolbar.h>
#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kcombobox.h>
#include <kpushbutton.h>
#include <kmainwindow.h>

extern int errno;

using namespace std;

KPS::KPS( QWidget *parent ) : KXmlGuiWindow( parent ),
						 updTime( 3000 ),
						 fTB( "filterToolBar" )
{
	// initialise the main view
	setCaption( i18n( "Process Status" ), false );

	// then, setup our actions
	setupActions();

	// and a status bar
	m_SBtext2 = new QLabel( i18n(" KPS initialising... "), this );
	m_SBtext2->setFrameShape( QFrame::StyledPanel );
	m_SBtext2->setFrameShadow( QFrame::Sunken );
	statusBar()->addWidget(m_SBtext2);

	m_SBtext1 = new QLabel( i18n(""), this );
	m_SBtext1->setFrameShape( QFrame::StyledPanel );
	m_SBtext1->setFrameShadow( QFrame::Sunken );
	statusBar()->addWidget(m_SBtext1);

	statusBar()->show();
	toolBar()->show();
	toolBar( fTB )->show();
	
	// build main widget
	initView();

	// apply the saved mainwindow settings, if any, and ask the mainwindow
	// to automatically save settings if changed: window size, toolbar
	// position, icon size, etc.
	setAutoSaveSettings();

	// set up timer to refresh main window
	m_update = new QTimer( this );				
	connect( m_update, SIGNAL(timeout()), this, SLOT(slotUpdateTimeout()) );
	m_update->start( updTime );
	
	m_SavePID = ""; 
	m_SaveName = "";
	filter = "";
	filtcol = ARG_COL;
	nullpx = new QPixmap( m_il.loadIcon( "unknownapp", KIconLoader::Small ) );

	setpriority(PRIO_PROCESS, 0, +10);          // lower our priority
	slotUpdateTimeout( true );									// kick things off
}

KPS::~KPS()
{
	m_update->stop();
	delete m_view;
}

void KPS::initView()
{
	// tell the KMainWindow that this is indeed the main widget
	m_view = new KPSView( this );
	setCentralWidget(m_view);
	m_view->show();
	toolBar( fTB )->show();
	m_TreeMode = false;
	
	// context menu for processes
	QAction *action;
	m_popup = new QMenu( m_view );
  action = m_popup->addAction( KIcon("view-media-equalizer"), QString(i18n("&Renice...")) );
	action->setData( 100 );
  m_popup->addSeparator();
  m_subpopup = m_popup->addMenu( KIcon("edit-delete"), QString(i18n("&Send signal")) );
	
	for ( int k = 1; k <= NUMSIGNALS; k++ )
	{
		action = m_subpopup->addAction( signaltypes[k].name );
		action->setData( k );
	}

  connect( m_view, SIGNAL(customContextMenuRequested(const QPoint &)),
     	     this, SLOT(slotRMBItem(const QPoint &)) );
  connect( m_view, SIGNAL(clicked(const QModelIndex &)),
     	     this, SLOT(slotRowChanged(const QModelIndex &)) );

	QString info = i18n(
		"<h3>Terms used on the process status display</h3>"
		"For more information use <b>man ps</b> or <b>man proc</b>.<br>"
		"<h4>Column headings</h4>"
		"<p><pre>"
		"PID   -    process identifier<br>"
		"%CPU  -    percentage of CPU time over real time<br>"
		"%MEM  -    percentage of memory used<br>"
		"VSZ   -    virtual machine memory size<br>"
		"RSS   -    resident set size (pages in memory)<br>"
		"Nice  -    priority (-19:19), lower value is higher<br>"
		"</pre></p>" 
		"<h4>Explanation of the Status column</h4>"
		"<p><pre>"
    "D     -    uninterruptible sleep<br>"
    "R     -    running<br>"
    "S     -    sleeping<br>"
    "T     -    traced or stopped<br>"
    "Z     -    defunct (zombie) process<br>"
    "</pre></p>"
    "<p><pre>"
    "W     -    has no resident pages<br>"
    "&lt;     -    high-priority task<br>"
    "N     -    low-priority task<br>"
    "L     -    has pages locked into memory<br>"
		"s     -    is a session leader<br>" 
		"l     -    is multi-threaded (using CLONE_THREAD, like NPTL pthreads do)<br>" 
		"+     -    is in the foreground process group<br>"
    "</pre></p>");
		
	m_view->setWhatsThis( info );
}

void KPS::setupActions()
{
	fileStart = newAction( "&Start", KIcon("media-playback-start"), Qt::CTRL+Qt::Key_S, this, "start" ); 
 	connect(fileStart, SIGNAL(triggered(bool)), this, SLOT(slotStart())); 
	fileStop = newAction( "Sto&p", KIcon("media-playback-stop"), Qt::CTRL+Qt::Key_P, this, "stop" );
	connect(fileStop, SIGNAL(triggered(bool)), this, SLOT(slotStop())); 
	fileReset = newAction( "Res&et", KIcon("view-refresh"), Qt::CTRL+Qt::Key_E, this, "reset" );
	connect(fileReset, SIGNAL(triggered(bool)), this, SLOT(slotReset())); 
	fileTree = newAction( "&Toggle tree View", KIcon("view-process-all-tree"), Qt::CTRL+Qt::Key_T, this, "tree" );
	connect(fileTree, SIGNAL(triggered(bool)), this, SLOT(slotTreeToggle())); 
	fileKill = newAction( "&Kill", KIcon("edit-delete"), Qt::Key_Delete, this, "kill" );
	connect(fileKill, SIGNAL(triggered(bool)), this, SLOT(slotKillItem())); 
	fileRenice = newAction( "&Renice...", KIcon("view-media-equalizer"), Qt::CTRL+Qt::Key_R, this, "renice" );
	connect(fileRenice, SIGNAL(triggered(bool)), this, SLOT(slotRenice())); 

	KStandardAction::quit(this, SLOT(slotQuit()), actionCollection());
	
	fileStart->setToolTip(i18n("Start updating status"));
	fileStop->setToolTip(i18n("Stop updating status"));
	fileKill->setToolTip(i18n("Kill selected process"));
	fileRenice->setToolTip(i18n("Change the priority of the selected process"));
	fileReset->setToolTip(i18n("Reset column sizes"));
	fileTree->setToolTip(i18n("Toggle process tree view"));
	
	// set up a secondary toolbar for filtering processes

	filterClear = new KAction( KIcon("edit-clear-history"), i18n(""), this );
	filterClear->setToolTip( i18n("Clear filter history") );
	actionCollection()->addAction( "filter_clear", filterClear );

	cb1 = new KComboBox( true, this );
	cb1->setToolTip( i18n("Type text or regexp to filter with and press Enter") );
	cb1->setMinimumWidth( 150 );
	cb1->clear();
	cb1->setDuplicatesEnabled( false );
	cb1->setInsertPolicy( QComboBox::InsertAtTop );

	le = new KLineEdit( cb1 );
	le->setClearButtonShown( true );
	le->setClickMessage( i18n("Filter") );
	cb1->setLineEdit( le );

	filterEdit = new KAction( i18n(""), this );
 	filterEdit->setDefaultWidget( cb1 );
	actionCollection()->addAction( "filter_edit", filterEdit );
	
	QStringList names;
	names << "Command" << "User" << "Status" <<  "Arguments";

	cb2 = new KComboBox( this );
	cb2->setToolTip( i18n("Choose column to filter with") );
	cb2->insertItems( 0, names );
	cb2->setCurrentIndex( 3 );

	filterNames = new KAction( i18n(""), this );
 	filterNames->setDefaultWidget( cb2 );
	actionCollection()->addAction( "filter_names", filterNames );

	connect( filterClear, SIGNAL(triggered(bool)), this, SLOT(slotEraseFilters()) ); 
	connect( cb1, SIGNAL(activated(const QString &)), this, SLOT(slotComboEdit(const QString &)) ); 
	connect( le, SIGNAL(clearButtonClicked()), this, SLOT(slotEraseEdit()) ); 
	connect( cb2, SIGNAL(highlighted(int)), this, SLOT(slotCombo(int)));
	
	setupGUI();   // build menu and tool bars
}

KAction* KPS::newAction(const QString &text, const KIcon &icon, const QKeySequence &sc, QObject *parent, const QString &xml_text )
{
	KAction *action = new KAction( icon, text, parent ); 
	action->setShortcut( sc );
	actionCollection()->addAction( xml_text, action );
	return action;
}

void KPS::changeStatusbar(const QString &text1, const QString &text2 )
{
	m_SBtext1->setText( text1 );
	m_SBtext2->setText( text2 );
}

void KPS::changeCaption(const QString& text)
{
  setCaption(text);
}

void KPS::slotUpdateTimeout( bool first )
{
	QByteArray state;
	
	if ( !first )
		state = m_view->header()->saveState();
	m_view->clear();
	
	if ( !m_TreeMode )
		initPSvalues();
	else
		initTree();

	if ( !first )
		m_view->header()->restoreState( state );
	
	changeStatusbar( QString( i18n("%1 processes: %2 sleeping, %3 running, %4 stopped, %5 zombie")
		 ).arg(stat_ALL).arg(stat_S).arg(stat_R).arg(stat_T).arg(stat_Z), getUptime() );
}

void KPS::slotQuit()
{
	close();
}

void KPS::slotRMBItem( const QPoint &point )
{
	slotRowChanged( QModelIndex() );

	QAction *action;
	action = m_popup->exec( m_view->viewport()->mapToGlobal(point) );    
	
	if ( action == 0 )
		return;
	else if ( action->data().toInt() == 100 )
		slotRenice();
	else
		slotSignalItem( action );
}

void KPS::slotRowChanged( const QModelIndex & )
{
	QTreeWidgetItem *item = m_view->currentItem();
	m_SavePID  = item->text( PID_COL ).trimmed(),
	m_SaveName = item->text( CMD_COL ).trimmed();
	m_SavePrio = item->text( NICE_COL ).trimmed();
}

void KPS::slotKillItem()
{
	slotSignalItem( 0 );
}

void KPS::slotSignalItem( QAction *action )
{
	if ( m_SavePID == "" ) return;

	const int signal = action == 0 ? 9 : action->data().toInt(); 
	
	const int res = KMessageBox::warningContinueCancel(this, 
									i18n("Request to send signal ") + signaltypes[signal].name + i18n(" to process ") + 
									m_SaveName + " (" + m_SavePID + ").\n" + i18n("Are you absolutely sure?"),
									i18n( "Kill Confirmation"));		
										
	if ( res == KMessageBox::Continue )
	{
		if ( kill( m_SavePID.toInt(), signal ) < 0 )
		{
			QString msg = strerror( errno );
			KMessageBox::error( this, msg, i18n("Kill Error") );
		}
		else
			slotUpdateTimeout();
	}
	m_SavePID = ""; m_SaveName = "";
}

void KPS::slotRenice()
{
	if ( m_SavePID == "" ) return;
	
	int prio = m_SavePrio.toInt();
	Renice *rnDialog = new Renice( this, m_SaveName, m_SavePID, &prio );
	const int rc = rnDialog->exec();
	delete rnDialog;
	
	if ( rc == QDialog::Accepted )
	{
		if ( setpriority(PRIO_PROCESS, m_SavePID.toInt(), prio) < 0 )
		{
			QString msg = strerror( errno );
			KMessageBox::error( this, msg, i18n("Renice Error") );
		}
		else
			slotUpdateTimeout();
	}
	m_SavePID = ""; m_SaveName = "";
}

void KPS::slotStart()
{
	if ( !m_update->isActive() )
	{
		m_update->start( updTime );
		changeStatusbar( i18n("Update restarted"), getUptime() );
	}
}

void KPS::slotStop()
{
	if ( m_update->isActive() )
	{
		m_update->stop();
		changeStatusbar( i18n("Update stopped"), getUptime() );
	}
}

void KPS::slotReset()
{
	// recreate the main view widget
	delete m_view;
	initView();
	slotUpdateTimeout( true );
}

void KPS::slotComboEdit( const QString &text )
{
	filter = text;
	slotUpdateTimeout();
}

void KPS::slotCombo( int col )
{
	switch  (col )
	{
		case 0: filtcol = CMD_COL;
			break;
		case 1: filtcol = USER_COL;
			break;
		case 2: filtcol = STAT_COL;
			break;
		case 3: filtcol = ARG_COL;
			break;
	}
}

void KPS::slotEraseEdit()
{
	filter = "";
	slotUpdateTimeout();
}

void KPS::slotEraseFilters()
{
	filter = "";
	le->clear();
	cb1->clear();
	slotUpdateTimeout();
}

void KPS::slotTreeToggle()
{
	if ( m_TreeMode )
	{
		m_view->setSortingEnabled( true );								// disable sorting
		m_view->sortItems( m_SaveSC, Qt::DescendingOrder );
		m_view->setRootIsDecorated( false );
		if ( !fTBhidden )
			toolBar( fTB )->show();
	}
	else
	{
		m_SaveSC = m_view->sortColumn();
		m_view->setSortingEnabled( false );								// disable sorting
		m_view->setRootIsDecorated( true );
		fTBhidden = toolBar( fTB )->isHidden();
		toolBar( fTB )->hide();
	}
	m_TreeMode = !m_TreeMode;
	slotUpdateTimeout();
}

void KPS::initPSvalues()
{
	QString tmp, cmd; int k;
	stat_S = stat_R = stat_T = stat_Z = stat_ALL = 0;

	// run ps command in pipe
	cmd = "ps h -eo " + QString( outFields );
	fp = popen( cmd.toAscii(), "r" );	
	
	while ( fgets( line, sizeof( line ), fp ) )
  {
		tmp = QString( line );
		tmp = tmp.remove( "<defunct>" );					// extra column for zombie processes
		tmp = tmp.simplified();    	
		
		// check we have all the line
		while ( !strchr( line, '\n' ) )
			fgets ( line, sizeof( line ), fp );     // read until newline found
			
		for ( k = CMD_COL; k < ARG_COL; ++k )
		{
			cols[k] = tmp.section( ' ', k, k );   // consider data as words separated by spaces
		}
		cols[ARG_COL] = tmp.section( ' ', ARG_COL );        // the rest
		
		stat_ALL += 1;
		if ( cols[STAT_COL].indexOf('S') >= 0 ) stat_S += 1;
		if ( cols[STAT_COL].indexOf('D') >= 0 ) stat_S += 1;
		if ( cols[STAT_COL].indexOf('R') >= 0 ) stat_R += 1;
		if ( cols[STAT_COL].indexOf('T') >= 0 ) stat_T += 1;
		if ( cols[STAT_COL].indexOf('Z') >= 0 ) stat_Z += 1;
		
		if ( filter.isEmpty() | cols[filtcol].indexOf( QRegExp(filter) ) >= 0 )
		{
			KPSListViewItem *item = new KPSListViewItem( m_view );
			QPixmap px;
			const int i = m_isNotApp.indexOf(cols[CMD_COL]);
			const int j = m_isApp.indexOf(cols[CMD_COL]);
			
			if ( i > -1 )
				item->setIcon( CMD_COL, QIcon(*nullpx) );
			else if ( j > -1 )
				item->setIcon( CMD_COL, QIcon(m_pxStor[j]) );
			else
			{
				px = m_il.loadIcon( cols[CMD_COL], KIconLoader::Small, 0, 
									KIconLoader::DefaultState, QStringList(), 0L, true );
				if ( QIcon(px).isNull() )
				{
					item->setIcon( CMD_COL, QIcon(*nullpx) );
					m_isNotApp << cols[CMD_COL];
				}
				else
				{
					item->setIcon( CMD_COL, QIcon(px) );
					m_isApp << cols[CMD_COL];
					m_pxStor << px;
				}
			}
			
			for ( k = CMD_COL; k <= ARG_COL; ++k )
			{
				item->setText( k, cols[k] ); 
			}
			
			item->setToolTip( ARG_COL, cols[ARG_COL] );
		}
	}                            
  pclose(fp);                 

	for ( k = CMD_COL; k <= ARG_COL; ++k )
	{
		m_view->resizeColumnToContents( k );
	}
}                      

void KPS::initTree()
{
	QString cmd;
	stat_S = stat_R = stat_T = stat_Z = stat_ALL = 0;
  
	// run ps command in pipe
	cmd = "ps h -eHo " + QString( outFields );
	fp = popen( cmd.toAscii(), "r" );	
	addItem( m_view );		  
	pclose( fp );

	for ( int k = CMD_COL; k <= ARG_COL; ++k )
	{
		m_view->resizeColumnToContents( k );
	}
}       

int KPS::addItem( QTreeWidget* parent )
{
	// level 0 is init (and maybe kthreadd) and is chained off the main view
	int indent = getLine();
 
	while ( indent == 0 ) 
	{                      
		KPSListViewItem *item = new KPSListViewItem( parent );               
		item->setExpanded( true );
		for ( int k = CMD_COL; k <= ARG_COL; ++k )
			item->setText( k, cols[k] ); 
		
		indent = addItem( item, getLine() );
	}
	return 0;
}

int KPS::addItem( KPSListViewItem* parent, const int level )
{
	KPSListViewItem *item = 0;               
	int indent = level;									// add current item at this level                      
	
	while ( indent >= level )                    
	{                                            
		if ( indent > level )                      
			indent = addItem( item, indent );        
		else                                       
		{                                          
			item = new KPSListViewItem( parent );                 
			item->setExpanded( true );
			for ( int k = CMD_COL; k <= ARG_COL; ++k )
				item->setText( k, cols[k] ); 
			indent = getLine();                      
		}                                          
	}                                            
	return indent;                               
}

int KPS::getLine()
{
	if ( !fgets( line, sizeof( line ), fp ) )
		return -1;														// end of file
		
	QString tmp( line ); int k;
	const int indent = tmp.indexOf( QRegExp( "\\S" ) );  // find first non-space 
	tmp = tmp.remove("<defunct>");					// extra column for zombie processes
	tmp = tmp.simplified();    	
	
	for ( k = CMD_COL; k < ARG_COL; ++k )
	{
		cols[k] = tmp.section( ' ', k, k );   // consider data as words separated by spaces
	}
	cols[ARG_COL] = tmp.section( ' ', ARG_COL );        // the rest
	
	stat_ALL += 1;
	if ( cols[STAT_COL].indexOf('S') >= 0 ) stat_S += 1;
	if ( cols[STAT_COL].indexOf('D') >= 0 ) stat_S += 1;
	if ( cols[STAT_COL].indexOf('R') >= 0 ) stat_R += 1;
	if ( cols[STAT_COL].indexOf('T') >= 0 ) stat_T += 1;
	if ( cols[STAT_COL].indexOf('Z') >= 0 ) stat_Z += 1;
	
	return indent;
}

QString KPS::getUptime()
{
	QString tmp, uptime;
	// get uptime like top
	fp = popen( "uptime", "r" );
	fgets( line, sizeof( line ), fp );
	tmp = line;
	const int m = tmp.indexOf("up"), n = tmp.indexOf("user");
	uptime = tmp.mid( m, n + 5 - m);		
  pclose(fp);        
	return uptime;         
}
                              
const KPS::t_signaltypes KPS::signaltypes[] =
     {
      { 0,  "ZERO" },
      { 1,  "SIGHUP" },
      { 2,  "SIGINT" },
      { 3,  "SIGQUIT" },
      { 4,  "SIGILL" },
      { 5,  "SIGTRAP" },
      { 6,  "SIGABRT" },
      { 7,  "SIGBUS" },
      { 8,  "SIGHFPE" },
      { 9,  "SIGKILL" },
      { 10, "SIGUSR1" },
      { 11, "SIGSEGV" },
      { 12, "SIGUSR2" },
      { 13, "SIGPIPE" },
      { 14, "SIGALRM" },
      { 15, "SIGTERM" },
      { 16, "SIGSTKFLT" },
      { 17, "SIGCHLD" },
      { 18, "SIGCONT" },
      { 19, "SIGSTOP" },
     };
                                     
#include "kps.moc"   
