/***************************************************************************
 *   Copyright (C) 2004 by Robin Atwood                                    *
 *   robin.atwood@attglobal.net                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _KPS_H_
#define _KPS_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qtimer.h>
#include <qlabel.h>
#include <qkeysequence.h>
#include <qmenu.h>
#include <qaction.h>
#include <QModelIndex>

#include <kapplication.h>
#include <kxmlguiwindow.h>
#include <kcompletion.h>
#include <kcombobox.h>
#include <kaction.h>
#include <kactioncollection.h>
#include <klineedit.h>

#include <sys/resource.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>

#include "kpsview.h"
#include "kpslistviewitem.h"
#include "renice.h"

class KPrinter;
class KToggleAction;
class KURL;

/**
 * This class serves as the main window for KPS.  It handles the
 * menus, toolbars, and status bars.
 *
 * @short Main window class
 * @author Robin Atwood <robin.atwood@attglobal.net>
 * @version 0.1
 */
class KPS : public KXmlGuiWindow
{
    Q_OBJECT
public:
    /**
     * Default Constructor
     */
    KPS( QWidget *parent=0 );

    /**
     * Default Destructor
     */
    virtual ~KPS();

protected:
    /**
     * Overridden virtuals for Qt drag 'n drop (XDND)
     */

protected:
		QTimer *m_update;
		QMenu *m_popup, *m_subpopup;
		
    static const int NUMSIGNALS = 19;
    typedef struct t_signaltypes{ int signal; QString name; };
    static const t_signaltypes signaltypes[NUMSIGNALS + 1];

		
    
private slots:
		KAction *newAction(const QString &, const KIcon &, const QKeySequence &, QObject *, const QString & );

    void changeStatusbar(const QString &text1, const QString &text2 );
    void changeCaption(const QString& text);

		void slotQuit();
		void slotStart();
		void slotStop();
		void slotReset();
		void slotUpdateTimeout( bool first = false );
		void slotKillItem();
		void slotSignalItem( QAction * );
		void slotRMBItem( const QPoint &point );
		void slotRowChanged( const QModelIndex & );
		void slotRenice();
		void slotEraseEdit();
		void slotEraseFilters();
		void slotCombo( int );
		void slotComboEdit( const QString &);
		void slotTreeToggle();
		
private:
		void initView();
    void setupAccel();
    void setupActions();
		void initPSvalues();		
		void initTree();
		int  addItem( QTreeWidget* );
		int  addItem( KPSListViewItem*, const int );
		int  getLine();
		QString getUptime();

private:
    KPSView			*m_view;
		KComboBox		*cb1, *cb2;
		KLineEdit		*le;
		KCompletion	*m_comp;
		KIconLoader m_il; 

		QLabel *m_SBtext1, *m_SBtext2;
		KAction *fileStart;
		KAction *fileStop;
		KAction *fileKill;
		KAction *fileReset;
		KAction *fileRenice;
		KAction *fileTree;
		KAction *fileQuit;
		KAction *filterClear;
		KAction *filterEdit;
		KAction *filterNames;

    KToggleAction *m_statusbarAction;
		
		QString m_SavePID;
		QString m_SaveName;
		QString m_SavePrio;
		int	 m_SaveSC;		
		QString cols[ARG_COL+1];	
		
		const int updTime;              /* update in msecs */
		int stat_S;
		int stat_R;
		int stat_T;
		int stat_Z;
		int stat_ALL;
	  char line[2048];										// line of data from unix command
	  FILE *fp;  
				
		QString filter;
		const QString fTB;
		int filtcol;
		bool m_TreeMode;
		bool fTBhidden;
		QPixmap *nullpx;
		QStringList m_isNotApp, m_isApp;
		QList<QPixmap> m_pxStor;
};

#endif // _KPS_H_
